@@ -9,7 +9,7 @@ fi
 VERSION=$1
 TAG=RELEASE_`echo $VERSION | sed -e 's/\./_/g'`
 
-dest=gabe.freedesktop.org:/home/srv/cairo.freedesktop.org/www/releases/
+dest=annarchy.freedesktop.org:/srv/cairo.freedesktop.org/www/releases/
 
 base=rcairo-$VERSION
 tgz=$base.tar.gz
@@ -18,6 +18,7 @@ tgz=$base.tar.gz
 cvs tag $TAG
 mkdir -p tmp
 cvs export -r $TAG -d tmp/$base rcairo
+cp -rp doc tmp/$base/
 tar cvfz $tgz -C tmp $base
 md5sum $tgz > $tgz.md5
 sha1sum $tgz > $tgz.sha1
