@@ -147,19 +147,19 @@ def pac(surface, width, height)
   cr.show_page
 end
 
-width = 841.889763779528
-height = 595.275590551181
+paper = Cairo::Paper.parse(:a4_landscape)
 
-Cairo::ImageSurface.new(width, height) do |surface|
-  cr = pac(surface, width, height)
+size_in_points = paper.size("pt")
+Cairo::ImageSurface.new(*size_in_points) do |surface|
+  cr = pac(surface, *size_in_points)
   cr.target.write_to_png("pac2.png")
 end
 
 scalable_surface_output = Proc.new do |surface_class_name, suffix|
   if Cairo.const_defined?(surface_class_name)
     surface_class = Cairo.const_get(surface_class_name)
-    surface_class.new("pac2.#{suffix}", width, height) do |surface|
-      pac(surface, width, height)
+    surface_class.new("pac2.#{suffix}", paper) do |surface|
+      pac(surface, *size_in_points)
     end
   else
     puts("#{surface_class_name} isn't supported.")
