@@ -3,7 +3,7 @@
  * Ruby Cairo Binding
  *
  * $Author: kou $
- * $Date: 2008-01-11 12:19:44 $
+ * $Date: 2008-08-16 12:52:17 $
  *
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
@@ -15,9 +15,34 @@
 #include "rb_cairo.h"
 #include "rb_cairo_private.h"
 
-#include "rubyio.h"
+#ifdef HAVE_RUBY_ST_H
+#  include <ruby/st.h>
+#else
+#  include <st.h>
+#endif
 
-#if CAIRO_HAS_PS_SURFACE || CAIRO_HAS_PDF_SURFACE || CAIRO_HAS_SVG_SURFACE
+#ifdef CAIRO_HAS_WIN32_SURFACE
+#  define OpenFile OpenFile_win32
+#  include <cairo-win32.h>
+#  undef OpenFile
+#endif
+
+#include <rubyio.h>
+
+#ifdef CAIRO_HAS_QUARTZ_SURFACE
+#  ifndef HAVE_TYPE_ENUM_RUBY_VALUE_TYPE
+enum ruby_value_type {
+  RUBY_T_DATA = T_DATA
+};
+#  endif
+#  undef T_DATA
+#  include <cairo-quartz.h>
+#  define T_DATA RUBY_T_DATA
+#endif
+
+#if defined(CAIRO_HAS_PS_SURFACE) || \
+  defined(CAIRO_HAS_PDF_SURFACE) || \
+  defined(CAIRO_HAS_SVG_SURFACE)
 #  define HAS_CREATE_CR_CLOSURE_SURFACE 1
 #else
 #  define HAS_CREATE_CR_CLOSURE_SURFACE 0
@@ -30,15 +55,40 @@ VALUE rb_cCairo_PDFSurface = Qnil;
 VALUE rb_cCairo_PSSurface = Qnil;
 VALUE rb_cCairo_SVGSurface = Qnil;
 VALUE rb_cCairo_Win32Surface = Qnil;
+VALUE rb_cCairo_Win32PrintingSurface = Qnil;
 VALUE rb_cCairo_QuartzSurface = Qnil;
+VALUE rb_cCairo_QuartzImageSurface = Qnil;
 
 static ID cr_id_target;
 static ID cr_id_read;
 static ID cr_id_write;
+static ID cr_id_inspect;
+static ID cr_id_parse;
+static ID cr_id_size;
+static ID cr_id_set_unit;
 static cairo_user_data_key_t cr_closure_key;
+static cairo_user_data_key_t cr_object_holder_key;
+static cairo_user_data_key_t cr_finished_key;
 
 #define _SELF  (RVAL2CRSURFACE(self))
 
+static VALUE
+cr_paper_parse (VALUE paper_description)
+{
+  return rb_funcall (rb_cCairo_Paper, cr_id_parse, 2, paper_description, Qtrue);
+}
+
+static void
+cr_paper_to_size_in_points (VALUE paper_description, VALUE *width, VALUE *height)
+{
+  VALUE paper, size;
+
+  paper = cr_paper_parse (paper_description);
+  size = rb_funcall (paper, cr_id_size, 1, rb_str_new2 ("pt"));
+  *width = RARRAY_PTR (size)[0];
+  *height = RARRAY_PTR (size)[1];
+}
+
 static inline void
 cr_surface_check_status (cairo_surface_t *surface)
 {
@@ -63,20 +113,44 @@ cr_surface_get_klass (cairo_surface_t *surface)
     case CAIRO_SURFACE_TYPE_PS:
       klass = rb_cCairo_PSSurface;
       break;
-    case CAIRO_SURFACE_TYPE_SVG:
-      klass = rb_cCairo_SVGSurface;
+    case CAIRO_SURFACE_TYPE_QUARTZ:
+      klass = rb_cCairo_QuartzSurface;
       break;
     case CAIRO_SURFACE_TYPE_WIN32:
       klass = rb_cCairo_Win32Surface;
       break;
+    case CAIRO_SURFACE_TYPE_SVG:
+      klass = rb_cCairo_SVGSurface;
+      break;
+#if CAIRO_CHECK_VERSION(1, 5, 2)
+    case CAIRO_SURFACE_TYPE_WIN32_PRINTING:
+      klass = rb_cCairo_Win32PrintingSurface;
+      break;
+#endif
+#if CAIRO_CHECK_VERSION(1, 5, 12)
+    case CAIRO_SURFACE_TYPE_QUARTZ_IMAGE:
+      klass = rb_cCairo_QuartzImageSurface;
+      break;
+#endif
     default:
-      rb_raise (rb_eArgError, "unknown source type: %d", type);
+      klass = rb_cCairo_Surface;
       break;
     }
 
+  if (NIL_P (klass))
+    rb_raise (rb_eArgError, "unknown source type: %d", type);
+
   return klass;
 }
 
+static char *
+inspect (VALUE object)
+{
+  VALUE inspected;
+  inspected = rb_funcall (object, cr_id_inspect, 0);
+  return StringValueCStr(inspected);
+}
+
 /* read/write callback */
 typedef struct cr_io_callback_closure {
   VALUE target;
@@ -85,6 +159,11 @@ typedef struct cr_io_callback_closure {
   unsigned int length;
 } cr_io_callback_closure_t;
 
+typedef struct cr_invoke_data {
+  cr_callback_func_t func;
+  VALUE data;
+} cr_invoke_data_t;
+
 #if HAS_CREATE_CR_CLOSURE_SURFACE
 static cr_io_callback_closure_t *
 cr_closure_new (VALUE target)
@@ -101,7 +180,7 @@ cr_closure_new (VALUE target)
 static void
 cr_closure_destroy (cr_io_callback_closure_t *closure)
 {
-  free (closure);
+  xfree (closure);
 }
 
 static void
@@ -120,6 +199,17 @@ cr_surface_io_func_rescue (VALUE io_closure)
   return Qnil;
 }
 
+static VALUE
+cr_surface_invoke_io_func (VALUE user_data)
+{
+  cr_invoke_data_t *data;
+
+  data = (cr_invoke_data_t *)user_data;
+  return rb_rescue2 (data->func, data->data,
+                     cr_surface_io_func_rescue, data->data, rb_eException,
+                     (VALUE)0);
+}
+
 /* write callback */
 static VALUE
 cr_surface_write_func_invoke (VALUE write_closure)
@@ -152,15 +242,16 @@ cr_surface_write_func (void *write_closure,
                        const unsigned char *data, unsigned int length)
 {
   cr_io_callback_closure_t *closure;
+  cr_invoke_data_t invoke_data;
 
   closure = (cr_io_callback_closure_t *)write_closure;
   closure->data = (unsigned char *)data;
   closure->length = length;
-  
-  rb_rescue2 (cr_surface_write_func_invoke, (VALUE) closure,
-              cr_surface_io_func_rescue, (VALUE) closure, rb_eException,
-              (VALUE)0);
-  
+
+  invoke_data.func = cr_surface_write_func_invoke;
+  invoke_data.data = (VALUE)closure;
+  rb_cairo__invoke_callback (cr_surface_invoke_io_func, (VALUE)&invoke_data);
+
   if (NIL_P (closure->error))
     return CAIRO_STATUS_SUCCESS;
   else
@@ -196,14 +287,16 @@ cr_surface_read_func (void *read_closure,
                       unsigned char *data, unsigned int length)
 {
   cr_io_callback_closure_t *closure;
+  cr_invoke_data_t invoke_data;
 
   closure = (cr_io_callback_closure_t *)read_closure;
   closure->data = data;
   closure->length = length;
-  rb_rescue2 (cr_surface_read_func_invoke, (VALUE) closure,
-              cr_surface_io_func_rescue, (VALUE) closure, rb_eException,
-              (VALUE)0);
-  
+
+  invoke_data.func = cr_surface_read_func_invoke;
+  invoke_data.data = (VALUE)closure;
+  rb_cairo__invoke_callback (cr_surface_invoke_io_func, (VALUE)&invoke_data);
+
   if (NIL_P (closure->error))
     return CAIRO_STATUS_SUCCESS;
   else
@@ -220,16 +313,30 @@ rb_cairo_surface_from_ruby_object (VALUE obj)
       rb_raise (rb_eTypeError, "not a cairo surface");
     }
   Data_Get_Struct (obj, cairo_surface_t, surface);
+  if (!surface)
+    rb_cairo_check_status (CAIRO_STATUS_NULL_POINTER);
   return surface;
 }
 
+static rb_cairo__object_holder_t *
+cr_object_holder_new (VALUE object)
+{
+  return rb_cairo__object_holder_new (rb_cCairo_Surface, object);
+}
+
+static void
+cr_object_holder_free (void *ptr)
+{
+  rb_cairo__object_holder_free (rb_cCairo_Surface, ptr);
+}
+
 static void
 cr_surface_free (void *ptr)
 {
-  if (ptr)
-    {
-      cairo_surface_destroy ((cairo_surface_t *) ptr);
-    }
+  cairo_surface_t *surface = ptr;
+
+  if (surface)
+    cairo_surface_destroy (surface);
 }
 
 VALUE
@@ -248,6 +355,18 @@ rb_cairo_surface_to_ruby_object (cairo_surface_t *surface)
     }
 }
 
+VALUE
+rb_cairo_surface_to_ruby_object_with_destroy (cairo_surface_t *surface)
+{
+  VALUE rb_surface;
+
+  rb_surface = rb_cairo_surface_to_ruby_object (surface);
+  if (surface)
+    cairo_surface_destroy (surface);
+
+  return rb_surface;
+}
+
 static VALUE
 cr_surface_allocate (VALUE klass)
 {
@@ -255,28 +374,52 @@ cr_surface_allocate (VALUE klass)
 }
 
 /* Surface manipulation */
+static VALUE
+cr_surface_destroy (VALUE self)
+{
+  cairo_surface_t *surface;
+
+  surface = _SELF;
+  cairo_surface_destroy (surface);
+  DATA_PTR (self) = NULL;
+
+  return self;
+}
+
+static VALUE
+cr_surface_finish (VALUE self)
+{
+  cairo_surface_t *surface;
+  cr_io_callback_closure_t *closure;
+
+  surface = _SELF;
+  closure = cairo_surface_get_user_data (surface, &cr_closure_key);
+
+  cairo_surface_finish (surface);
+  cairo_surface_set_user_data (surface, &cr_finished_key, (void *)CR_TRUE, NULL);
+  cairo_surface_set_user_data (surface, &cr_object_holder_key, NULL, NULL);
+
+  if (closure && !NIL_P (closure->error))
+    rb_exc_raise (closure->error);
+  cr_surface_check_status (surface);
+
+  return self;
+}
+
 static void
 yield_and_finish (VALUE self)
 {
   cairo_surface_t *surface;
-  cairo_status_t status;
 
   rb_yield (self);
 
   surface = _SELF;
-  if (cairo_surface_status (surface))
-    return;
-  cairo_surface_finish (surface);
-  status = cairo_surface_status (surface);
-  if (status == CAIRO_STATUS_SUCCESS || status == CAIRO_STATUS_SURFACE_FINISHED)
-    return;
-
-  cr_surface_check_status (surface);
+  if (!cairo_surface_get_user_data (surface, &cr_finished_key))
+    cr_surface_finish (self);
 }
 
 static VALUE
-cr_surface_create_similar (VALUE self, VALUE content,
-                           VALUE width, VALUE height)
+cr_surface_create_similar (VALUE self, VALUE content, VALUE width, VALUE height)
 {
   cairo_surface_t *surface;
 
@@ -284,22 +427,7 @@ cr_surface_create_similar (VALUE self, VALUE content,
                                           RVAL2CRCONTENT (content),
                                           NUM2INT (width), NUM2INT (height));
   cr_surface_check_status (surface);
-  return CRSURFACE2RVAL (surface);
-}
-
-static VALUE
-cr_surface_finish (VALUE self)
-{
-  cr_io_callback_closure_t *closure;
-  closure = cairo_surface_get_user_data (_SELF, &cr_closure_key);
-  
-  cairo_surface_finish (_SELF);
-
-  if (closure && !NIL_P (closure->error))
-    rb_exc_raise (closure->error);
-  
-  cr_surface_check_status (_SELF);
-  return self;
+  return CRSURFACE2RVAL_WITH_DESTROY (surface);
 }
 
 static VALUE
@@ -309,7 +437,7 @@ cr_surface_get_content (VALUE self)
 }
 
 
-#if CAIRO_HAS_PNG_FUNCTIONS
+#ifdef CAIRO_HAS_PNG_FUNCTIONS
 static VALUE
 cr_surface_write_to_png_stream (VALUE self, VALUE target)
 {
@@ -385,8 +513,18 @@ cr_surface_mark_dirty (int argc, VALUE *argv, VALUE self)
     }
   else
     {
+      int i;
+      VALUE args;
+
+      args = rb_ary_new2 (n);
+      for (i = 0; i < n; i++)
+        {
+          rb_ary_push (args, argv[i]);
+        }
+
       rb_raise (rb_eArgError,
-                "invalid argument (expect () or (x, y, width, height))");
+                "invalid argument (expect () or (x, y, width, height)): %s",
+                inspect (args));
     }
 
   cr_surface_check_status (_SELF);
@@ -425,9 +563,43 @@ cr_surface_set_fallback_resolution (VALUE self,
   return self;
 }
 
+#if CAIRO_CHECK_VERSION(1, 7, 2)
+static VALUE
+cr_surface_get_fallback_resolution (VALUE self)
+{
+  double x_pixels_per_inch, y_pixels_per_inch;
+
+  cairo_surface_get_fallback_resolution (_SELF,
+                                         &x_pixels_per_inch,
+                                         &y_pixels_per_inch);
+  cr_surface_check_status (_SELF);
+  return rb_ary_new3 (2,
+                      rb_float_new (x_pixels_per_inch),
+                      rb_float_new (y_pixels_per_inch));
+}
+#endif
+
+
+#if CAIRO_CHECK_VERSION(1, 5, 2)
+static VALUE
+cr_surface_copy_page (VALUE self)
+{
+  cairo_surface_copy_page (_SELF);
+  cr_surface_check_status (_SELF);
+  return self;
+}
+
+static VALUE
+cr_surface_show_page (VALUE self)
+{
+  cairo_surface_show_page (_SELF);
+  cr_surface_check_status (_SELF);
+  return self;
+}
+#endif
 
 /* Image-surface functions */
-#if CAIRO_HAS_PNG_FUNCTIONS
+#ifdef CAIRO_HAS_PNG_FUNCTIONS
 static cairo_surface_t *
 cr_image_surface_create_from_png_stream (VALUE target)
 {
@@ -469,7 +641,6 @@ cr_image_surface_create_from_png_generic (VALUE klass, VALUE target)
 }
 #endif
 
-
 static cairo_surface_t *
 cr_image_surface_create (VALUE self, VALUE format, VALUE width, VALUE height)
 {
@@ -516,7 +687,8 @@ cr_image_surface_initialize (int argc, VALUE *argv, VALUE self)
               "invalid argument (expect "
               "(width, height) or "
               "(format, width, height) or "
-              "(data, format, width, height, stride))");
+              "(data, format, width, height, stride)): %s",
+              inspect (rb_ary_new3 (4, arg1, arg2, arg3, arg4)));
 
   cr_surface_check_status (surface);
   DATA_PTR (self) = surface;
@@ -570,13 +742,29 @@ cr_image_surface_get_stride (VALUE self)
 /* Printing surfaces */
 #define DEFINE_SURFACE(type)                                            \
 static VALUE                                                            \
-cr_ ## type ## _surface_initialize (VALUE self, VALUE target,           \
-                                    VALUE rb_width_in_points,           \
-                                    VALUE rb_height_in_points)          \
+cr_ ## type ## _surface_initialize (int argc, VALUE *argv, VALUE self)  \
 {                                                                       \
+  VALUE target, rb_width_in_points, rb_height_in_points;                \
+  VALUE arg2, arg3;                                                     \
   cairo_surface_t *surface;                                             \
   double width_in_points, height_in_points;                             \
                                                                         \
+  rb_scan_args (argc, argv, "21", &target, &arg2, &arg3);               \
+  if (argc == 2)                                                        \
+    {                                                                   \
+      VALUE paper;                                                      \
+                                                                        \
+      paper = arg2;                                                     \
+      cr_paper_to_size_in_points (paper,                                \
+                                  &rb_width_in_points,                  \
+                                  &rb_height_in_points);                \
+    }                                                                   \
+  else                                                                  \
+    {                                                                   \
+      rb_width_in_points = arg2;                                        \
+      rb_height_in_points = arg3;                                       \
+    }                                                                   \
+                                                                        \
   width_in_points = NUM2DBL (rb_width_in_points);                       \
   height_in_points = NUM2DBL (rb_height_in_points);                     \
                                                                         \
@@ -601,6 +789,9 @@ cr_ ## type ## _surface_initialize (VALUE self, VALUE target,           \
           rb_ivar_set (self, cr_id_target, target);                     \
           cairo_surface_set_user_data (surface, &cr_closure_key,        \
                                        closure, cr_closure_free);       \
+          cairo_surface_set_user_data (surface, &cr_object_holder_key,  \
+                                       cr_object_holder_new(self),      \
+                                       cr_object_holder_free);          \
         }                                                               \
     }                                                                   \
   else                                                                  \
@@ -618,22 +809,40 @@ cr_ ## type ## _surface_initialize (VALUE self, VALUE target,           \
   return Qnil;                                                          \
 }
 
+#define DEFINE_SURFACE_SET_SIZE(type)                                   \
+static VALUE                                                            \
+cr_ ## type ## _surface_set_size (int argc, VALUE *argv, VALUE self)    \
+{                                                                       \
+  VALUE arg1, arg2;                                                     \
+  VALUE width_in_points, height_in_points;                              \
+                                                                        \
+  rb_scan_args(argc, argv, "11", &arg1, &arg2);                         \
+  if (argc == 1)                                                        \
+    {                                                                   \
+      VALUE paper;                                                      \
+                                                                        \
+      paper = arg1;                                                     \
+      cr_paper_to_size_in_points (paper,                                \
+                                  &width_in_points,                     \
+                                  &height_in_points);                   \
+    }                                                                   \
+  else                                                                  \
+    {                                                                   \
+      width_in_points = arg1;                                           \
+      height_in_points = arg2;                                          \
+    }                                                                   \
+                                                                        \
+  cairo_ ## type ## _surface_set_size (_SELF,                           \
+                                       NUM2DBL (width_in_points),       \
+                                       NUM2DBL (height_in_points));     \
+  cr_surface_check_status (_SELF);                                      \
+  return Qnil;                                                          \
+}
 
-#if CAIRO_HAS_PS_SURFACE
+#ifdef CAIRO_HAS_PS_SURFACE
 /* PS-surface functions */
 DEFINE_SURFACE(ps)
-
-static VALUE
-cr_ps_surface_set_size (VALUE self,
-                        VALUE width_in_points,
-                        VALUE height_in_points)
-{
-  cairo_ps_surface_set_size (_SELF,
-                             NUM2DBL (width_in_points),
-                             NUM2DBL (height_in_points));
-  cr_surface_check_status (_SELF);
-  return Qnil;
-}
+DEFINE_SURFACE_SET_SIZE(ps)
 
 static VALUE
 cr_ps_surface_dsc_comment (VALUE self, VALUE comment)
@@ -664,26 +873,39 @@ cr_ps_surface_dsc_begin_page_setup (VALUE self)
   else
     return Qnil;
 }
-#endif
-
-#if CAIRO_HAS_PDF_SURFACE
-/* PDF-surface functions */
-DEFINE_SURFACE(pdf)
 
+#  if CAIRO_CHECK_VERSION(1, 5, 2)
 static VALUE
-cr_pdf_surface_set_size (VALUE self,
-                         VALUE width_in_points,
-                         VALUE height_in_points)
+cr_ps_surface_restrict_to_level (VALUE self, VALUE level)
 {
-  cairo_pdf_surface_set_size (_SELF,
-                              NUM2DBL (width_in_points),
-                              NUM2DBL (height_in_points));
+  cairo_ps_surface_restrict_to_level (_SELF, RVAL2CRPSLEVEL (level));
   cr_surface_check_status (_SELF);
   return Qnil;
 }
+
+static VALUE
+cr_ps_surface_get_eps (VALUE self)
+{
+  return cairo_ps_surface_get_eps (_SELF) ? Qtrue : Qfalse;
+}
+
+static VALUE
+cr_ps_surface_set_eps (VALUE self, VALUE eps)
+{
+  cairo_ps_surface_set_eps (_SELF, RTEST (eps));
+  cr_surface_check_status (_SELF);
+  return Qnil;
+}
+#  endif
 #endif
 
-#if CAIRO_HAS_SVG_SURFACE
+#ifdef CAIRO_HAS_PDF_SURFACE
+/* PDF-surface functions */
+DEFINE_SURFACE(pdf)
+DEFINE_SURFACE_SET_SIZE(pdf)
+#endif
+
+#ifdef CAIRO_HAS_SVG_SURFACE
 /* SVG-surface functions */
 DEFINE_SURFACE(svg)
 
@@ -694,36 +916,9 @@ cr_svg_surface_restrict_to_version (VALUE self, VALUE version)
   cr_surface_check_status (_SELF);
   return Qnil;
 }
-
-static VALUE
-cr_svg_get_versions (VALUE self)
-{
-  VALUE rb_versions;
-  int i, num_versions;
-  cairo_svg_version_t const *versions;
-
-  cairo_svg_get_versions (&versions, &num_versions);
-
-  rb_versions = rb_ary_new2 (num_versions);
-
-  for (i = 0; i < num_versions; i++)
-    {
-      rb_ary_push (rb_versions, INT2NUM (versions[i]));
-    }
-
-  return rb_versions;
-}
-
-static VALUE
-cr_svg_version_to_string (VALUE self, VALUE version)
-{
-  const char *ver_str;
-  ver_str = cairo_svg_version_to_string (RVAL2CRSVGVERSION(version));
-  return rb_str_new2 (ver_str);
-}
 #endif
 
-#if CAIRO_HAS_WIN32_SURFACE
+#ifdef CAIRO_HAS_WIN32_SURFACE
 /* WIN32-surface functions */
 
 /* from dl/dl.h (ruby 1.9) */
@@ -762,14 +957,21 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
           (rb_cairo__is_kind_of (arg1, rb_cNumeric) &&
            NUM2INT (arg1) != CAIRO_FORMAT_RGB24))
         {
+#  if CAIRO_CHECK_VERSION(1, 4, 0)
+          HDC win32_hdc;
           hdc = arg1;
           width = arg2;
           height = arg3;
-          HDC win32_hdc = NIL_P (hdc) ? NULL : NUM2PTR (hdc);
+          win32_hdc = NIL_P (hdc) ? NULL : NUM2PTR (hdc);
           surface = cairo_win32_surface_create_with_ddb (win32_hdc,
                                                          CAIRO_FORMAT_RGB24,
                                                          NUM2INT (width),
                                                          NUM2INT (height));
+#  else
+          rb_raise (rb_eArgError,
+                    "Cairo::Win32Surface.new(hdc, width, height) "
+                    "is available since cairo >= 1.4.0");
+#  endif
         }
       else
         {
@@ -783,6 +985,7 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
       break;
     case 4:
       {
+#  if CAIRO_CHECK_VERSION(1, 4, 0)
         HDC win32_hdc;
         hdc = arg1;
         format = arg2;
@@ -793,6 +996,11 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
                                                        RVAL2CRFORMAT (format),
                                                        NUM2INT (width),
                                                        NUM2INT (height));
+#  else
+        rb_raise (rb_eArgError,
+                  "Cairo::Win32Surface.new(hdc, format, width, height) "
+                  "is available since cairo >= 1.4.0");
+#  endif
       }
       break;
     }
@@ -806,6 +1014,21 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
   return Qnil;
 }
 
+#  if CAIRO_CHECK_VERSION(1, 5, 2)
+static VALUE
+cr_win32_printing_surface_initialize (VALUE self, VALUE hdc)
+{
+  cairo_surface_t *surface = NULL;
+
+  surface = cairo_win32_printing_surface_create (NUM2PTR (hdc));
+  cr_surface_check_status (surface);
+  DATA_PTR (self) = surface;
+  if (rb_block_given_p ())
+    yield_and_finish (self);
+  return Qnil;
+}
+#  endif
+
 static VALUE
 cr_win32_surface_get_hdc (VALUE self)
 {
@@ -818,6 +1041,7 @@ cr_win32_surface_get_hdc (VALUE self)
     return PTR2NUM (hdc);
 }
 
+#  if CAIRO_CHECK_VERSION(1, 4, 0)
 static VALUE
 cr_win32_surface_get_image (VALUE self)
 {
@@ -826,19 +1050,19 @@ cr_win32_surface_get_image (VALUE self)
   surface = cairo_win32_surface_get_image (_SELF);
   if (!surface)
     return Qnil;
-  rb_cairo_check_status (cairo_surface_status (surface));
+  cr_surface_check_status (surface);
   return CRSURFACE2RVAL (surface);
 }
+#  endif
 #endif
 
-#if CAIRO_HAS_QUARTZ_SURFACE
+#if defined(CAIRO_HAS_QUARTZ_SURFACE) && defined(HAVE_RUBY_COCOA)
 /* Quartz-surface functions */
 
 #include <objc/objc-runtime.h>
 
-VALUE ocobj_s_new (id ocid);
-id rbobj_get_ocid (VALUE obj);
-
+BOOL rbobj_to_nsobj (VALUE obj, id* nsobj);
+VALUE ocid_to_rbobj (VALUE context_obj, id ocid);
 
 static VALUE
 cr_quartz_surface_initialize (int argc, VALUE *argv, VALUE self)
@@ -849,6 +1073,7 @@ cr_quartz_surface_initialize (int argc, VALUE *argv, VALUE self)
   cairo_surface_t *surface = NULL;
   cairo_format_t format = CAIRO_FORMAT_ARGB32;
   VALUE arg1, arg2, arg3, rb_width, rb_height;
+  static VALUE rb_cOSXCGContextRef = Qnil;
 
   rb_scan_args (argc, argv, "21", &arg1, &arg2, &arg3);
 
@@ -859,9 +1084,32 @@ cr_quartz_surface_initialize (int argc, VALUE *argv, VALUE self)
     }
   else
     {
-      objc_object = rbobj_get_ocid (arg1);
-      if (objc_object == nil)
-        format = RVAL2CRFORMAT (arg1);
+      switch (TYPE (arg1))
+        {
+        case T_NIL:
+          break;
+        case T_STRING:
+        case T_SYMBOL:
+        case T_FIXNUM:
+          format = RVAL2CRFORMAT (arg1);
+          break;
+        default:
+          if (NIL_P (rb_cOSXCGContextRef))
+            rb_cOSXCGContextRef =
+              rb_const_get (rb_const_get (rb_cObject, rb_intern ("OSX")),
+                            rb_intern ("CGContextRef"));
+
+          if (RTEST (rb_obj_is_kind_of (arg1, rb_cOSXCGContextRef)))
+            rbobj_to_nsobj (arg1, &objc_object);
+          else
+            rb_raise (rb_eArgError,
+                      "invalid argument (expect "
+                      "(width, height), "
+                      "(format, width, height) or "
+                      "(cg_context, width, height)): %s",
+                      inspect (rb_ary_new3 (3, arg1, arg2, arg3)));
+          break;
+        }
 
       rb_width = arg2;
       rb_height = arg3;
@@ -896,10 +1144,51 @@ cr_quartz_surface_get_cg_context (VALUE self)
 
   context = cairo_quartz_surface_get_cg_context (_SELF);
   objc_object = (id)context;
-  return ocobj_s_new (objc_object);
+  return ocid_to_rbobj (Qnil, objc_object);
 }
+
+#  if CAIRO_CHECK_VERSION(1, 5, 12)
+static VALUE
+cr_quartz_image_surface_initialize (VALUE self, VALUE image_surface)
+{
+  cairo_surface_t *surface;
+
+  surface = cairo_quartz_image_surface_create (RVAL2CRSURFACE (image_surface));
+  cr_surface_check_status (surface);
+  DATA_PTR (self) = surface;
+  if (rb_block_given_p ())
+    yield_and_finish (self);
+  return Qnil;
+}
+
+static VALUE
+cr_quartz_image_surface_get_image (VALUE self)
+{
+  cairo_surface_t *surface;
+
+  surface = cairo_quartz_image_surface_get_image (_SELF);
+  if (!surface)
+    return Qnil;
+  cr_surface_check_status (surface);
+  return CRSURFACE2RVAL (surface);
+}
+#  endif
 #endif
 
+static int
+cr_finish_all_guarded_surfaces_at_end_iter (VALUE key, VALUE value, VALUE data)
+{
+  cr_surface_finish (key);
+  return ST_CONTINUE;
+}
+
+static void
+cr_finish_all_guarded_surfaces_at_end (VALUE data)
+{
+  rb_hash_foreach (rb_cairo__gc_guarded_objects (rb_cCairo_Surface),
+                   cr_finish_all_guarded_surfaces_at_end_iter,
+                   Qnil);
+}
 
 void
 Init_cairo_surface (void)
@@ -907,28 +1196,47 @@ Init_cairo_surface (void)
   cr_id_target = rb_intern ("target");
   cr_id_read = rb_intern ("read");
   cr_id_write = rb_intern ("write");
-  
+  cr_id_inspect = rb_intern ("inspect");
+  cr_id_parse = rb_intern ("parse");
+  cr_id_size = rb_intern ("size");
+  cr_id_set_unit = rb_intern ("unit=");
+
   rb_cCairo_Surface =
     rb_define_class_under (rb_mCairo, "Surface", rb_cObject);
   rb_define_alloc_func (rb_cCairo_Surface, cr_surface_allocate);
-  
+
+  rb_cairo__initialize_gc_guard_holder_class (rb_cCairo_Surface);
+  rb_set_end_proc(cr_finish_all_guarded_surfaces_at_end, Qnil);
+
+
   rb_define_method (rb_cCairo_Surface, "create_similar",
                     cr_surface_create_similar, 3);
+  rb_define_method (rb_cCairo_Surface, "destroy", cr_surface_destroy, 0);
   rb_define_method (rb_cCairo_Surface, "finish", cr_surface_finish, 0);
   rb_define_method (rb_cCairo_Surface, "content", cr_surface_get_content, 0);
 
   rb_define_method (rb_cCairo_Surface, "font_options",
                     cr_surface_get_font_options, 0);
   rb_define_method (rb_cCairo_Surface, "flush", cr_surface_flush, 0);
-  rb_define_method (rb_cCairo_Surface, "mark_dirty", cr_surface_mark_dirty, 0);
+  rb_define_method (rb_cCairo_Surface, "mark_dirty", cr_surface_mark_dirty, -1);
   rb_define_method (rb_cCairo_Surface, "set_device_offset",
                     cr_surface_set_device_offset, 2);
   rb_define_method (rb_cCairo_Surface, "device_offset",
                     cr_surface_get_device_offset, 0);
   rb_define_method (rb_cCairo_Surface, "set_fallback_resolution",
                     cr_surface_set_fallback_resolution, 2);
+#if CAIRO_CHECK_VERSION(1, 7, 2)
+  rb_define_method (rb_cCairo_Surface, "fallback_resolution",
+                    cr_surface_get_fallback_resolution, 0);
+#endif
+#if CAIRO_CHECK_VERSION(1, 5, 2)
+  rb_define_method (rb_cCairo_Surface, "copy_page",
+                    cr_surface_copy_page, 2);
+  rb_define_method (rb_cCairo_Surface, "show_page",
+                    cr_surface_show_page, 2);
+#endif
 
-#if CAIRO_HAS_PNG_FUNCTIONS
+#ifdef CAIRO_HAS_PNG_FUNCTIONS
   rb_define_method (rb_cCairo_Surface, "write_to_png",
                     cr_surface_write_to_png_generic, 1);
 #endif
@@ -939,7 +1247,7 @@ Init_cairo_surface (void)
   rb_cCairo_ImageSurface =
     rb_define_class_under (rb_mCairo, "ImageSurface", rb_cCairo_Surface);
   
-#if CAIRO_HAS_PNG_FUNCTIONS
+#ifdef CAIRO_HAS_PNG_FUNCTIONS
   rb_define_singleton_method (rb_cCairo_ImageSurface, "from_png",
                               cr_image_surface_create_from_png_generic, 1);
 #endif
@@ -964,13 +1272,13 @@ Init_cairo_surface (void)
                            rb_cCairo_Surface);                          \
                                                                         \
   rb_define_method (rb_cCairo_ ## name ## Surface, "initialize",        \
-                    cr_ ## type ## _surface_initialize, 3);
+                    cr_ ## type ## _surface_initialize, -1);
 
-#if CAIRO_HAS_PS_SURFACE
+#ifdef CAIRO_HAS_PS_SURFACE
   /* PS-surface */
   INIT_SURFACE(ps, PS)
 
-  rb_define_method (rb_cCairo_PSSurface, "set_size", cr_ps_surface_set_size, 2);
+  rb_define_method (rb_cCairo_PSSurface, "set_size", cr_ps_surface_set_size, -1);
   rb_define_method (rb_cCairo_PSSurface, "dsc_comment",
                     cr_ps_surface_dsc_comment, 1);
   rb_define_method (rb_cCairo_PSSurface, "dsc_begin_setup",
@@ -978,35 +1286,37 @@ Init_cairo_surface (void)
   rb_define_method (rb_cCairo_PSSurface, "dsc_begin_page_setup",
                     cr_ps_surface_dsc_begin_page_setup, 0);
 
+#  if CAIRO_CHECK_VERSION(1, 5, 2)
+  rb_define_method (rb_cCairo_PSSurface, "restrict_to_level",
+                    cr_ps_surface_restrict_to_level, 1);
+  rb_define_method (rb_cCairo_PSSurface, "eps?", cr_ps_surface_get_eps, 0);
+  rb_define_method (rb_cCairo_PSSurface, "set_eps", cr_ps_surface_set_eps, 1);
+#  endif
+
   RB_CAIRO_DEF_SETTERS (rb_cCairo_PSSurface);
 #endif
 
-#if CAIRO_HAS_PDF_SURFACE
+#ifdef CAIRO_HAS_PDF_SURFACE
   /* PDF-surface */
   INIT_SURFACE(pdf, PDF)
 
   rb_define_method (rb_cCairo_PDFSurface, "set_size",
-                    cr_pdf_surface_set_size, 2);
+                    cr_pdf_surface_set_size, -1);
 
   RB_CAIRO_DEF_SETTERS (rb_cCairo_PDFSurface);
 #endif
 
-#if CAIRO_HAS_SVG_SURFACE
+#ifdef CAIRO_HAS_SVG_SURFACE
   /* SVG-surface */
   INIT_SURFACE(svg, SVG)
 
-  rb_define_singleton_method (rb_cCairo_SVGSurface, "versions",
-                              cr_svg_get_versions, 0);
-  rb_define_singleton_method (rb_cCairo_SVGSurface, "version_to_string",
-                              cr_svg_version_to_string, 1);
-
   rb_define_method (rb_cCairo_SVGSurface, "restrict_to_version",
                     cr_svg_surface_restrict_to_version, 1);
 
   RB_CAIRO_DEF_SETTERS (rb_cCairo_SVGSurface);
 #endif
 
-#if CAIRO_HAS_WIN32_SURFACE
+#ifdef CAIRO_HAS_WIN32_SURFACE
   /* Win32-surface */
   rb_cCairo_Win32Surface =
     rb_define_class_under (rb_mCairo, "Win32Surface", rb_cCairo_Surface);
@@ -1015,11 +1325,24 @@ Init_cairo_surface (void)
                     cr_win32_surface_initialize, -1);
   rb_define_method (rb_cCairo_Win32Surface, "hdc",
                     cr_win32_surface_get_hdc, 0);
+#  if CAIRO_CHECK_VERSION(1, 4, 0)
   rb_define_method (rb_cCairo_Win32Surface, "image",
                     cr_win32_surface_get_image, 0);
+#  endif
+
+#  if CAIRO_CHECK_VERSION(1, 5, 2)
+  rb_cCairo_Win32PrintingSurface =
+    rb_define_class_under (rb_mCairo, "Win32PrintingSurface", rb_cCairo_Surface);
+
+  rb_define_method (rb_cCairo_Win32PrintingSurface, "initialize",
+                    cr_win32_printing_surface_initialize, -1);
+  rb_define_method (rb_cCairo_Win32PrintingSurface, "hdc",
+                    cr_win32_surface_get_hdc, 0);
+#  endif
+
 #endif
 
-#if CAIRO_HAS_QUARTZ_SURFACE
+#if defined(CAIRO_HAS_QUARTZ_SURFACE) && defined(HAVE_RUBY_COCOA)
   /* Quartz-surface */
 
   rb_cCairo_QuartzSurface =
@@ -1029,5 +1352,16 @@ Init_cairo_surface (void)
                     cr_quartz_surface_initialize, -1);
   rb_define_method (rb_cCairo_QuartzSurface, "cg_context",
                     cr_quartz_surface_get_cg_context, 0);
+
+#  if CAIRO_CHECK_VERSION(1, 5, 12)
+  rb_cCairo_QuartzImageSurface =
+    rb_define_class_under (rb_mCairo, "QuartzImageSurface", rb_cCairo_Surface);
+
+  rb_define_method (rb_cCairo_QuartzImageSurface, "initialize",
+                    cr_quartz_image_surface_initialize, 1);
+  rb_define_method (rb_cCairo_QuartzImageSurface, "image",
+                    cr_quartz_image_surface_get_image, 0);
+#  endif
+
 #endif
 }
